

class Environment:
    def __init__(self):
        self.store = {}

    def get(self, name):
        obj = self.store.get(name)
        return obj

    def set(self, name, val):
        self.store[name] = val
        return val



from .token import Token, TokenType, lookup_keyword, is_color, is_property


class TokenUnKnownException(Exception):
    pass


def is_letter(ch):

    return 'a' <= ch <= 'z' or 'A' <= ch <= 'Z' or '0' <= ch <= '9' or ch in ['-', '_']


def is_digit(ch):
    return '0' <= ch <= '9'


def is_white_space(ch):
    return ch == ' ' or ch == '\t' or ch == '\n' or ch == '\r'


class Lexer:
    def __init__(self, input_str: str):
        self.input_str: str = input_str
        self.cur_pos: int = 0
        self.peek_pos: int = 0
        self.cur_char = None
        self.peek_char = None

        self.read_char()

    def read_char(self):
        if self.peek_pos >= len(self.input_str):
            self.cur_char = 0
        else:
            self.cur_char = self.input_str[self.peek_pos]
            self.cur_pos = self.peek_pos
            self.peek_pos += 1
            if not self.peek_pos >= len(self.input_str):
                self.peek_char = self.input_str[self.peek_pos]

        return self.cur_char

    # ':+(){}'
    def next_token(self) -> Token:

        self.skip_white_space()
        self.skip_comment()
        self.skip_white_space()

        lexeme = self.cur_char
        tok = None
        if lexeme == 0:
            tok = Token(TokenType.EOF, lexeme)
        elif lexeme == ':':
            tok = Token(TokenType.ASSIGN, lexeme)
        elif lexeme == '+':
            tok = Token(TokenType.PLUS, lexeme)
        elif lexeme == '-':
            tok = Token(TokenType.SUB, lexeme)
        elif lexeme == '*':
            tok = Token(TokenType.MUL, lexeme)
        elif lexeme == '/':
            tok = Token(TokenType.DIV, lexeme)
        elif lexeme == '>':
            tok = Token(TokenType.GT, lexeme)
        elif lexeme == '<':
            tok = Token(TokenType.LT, lexeme)
        elif lexeme == '=':
            lexeme = '=='
            self.read_char()
            tok = Token(TokenType.EQ, lexeme)
        elif lexeme == '(':
            tok = Token(TokenType.LEFT_PAREN, lexeme)
        elif lexeme == ')':
            tok = Token(TokenType.RIGHT_PAREN, lexeme)
        elif lexeme == '{':
            tok = Token(TokenType.LEFT_BRACE, lexeme)
        elif lexeme == '}':
            tok = Token(TokenType.RIGHT_BRACE, lexeme)
        elif lexeme == ';':
            tok = Token(TokenType.SEMICOLON, lexeme)
        elif lexeme == ',':
            tok = Token(TokenType.COMMA, lexeme)
        elif lexeme == '"' or lexeme == "'":
            pass  # 字符串 目前不接受插值
        elif lexeme == '&':
            tok = Token(TokenType.PARENT_REFERENCE, lexeme)
        elif is_digit(lexeme):
            lexeme = self.read_number()
            tok = Token(TokenType.NUMBER, lexeme)
        elif lexeme == '$':
            lexeme = self.read_identifier()
            tok = Token(TokenType.IDENTIFIER, lexeme)
        elif lexeme == '!' or lexeme == '@':
            if self.peek_char == '=':
                lexeme = '!='
                self.read_char()
                tok = Token(TokenType.EQ, lexeme)
            else:
                tok = self.read_keyword()
        elif lexeme == '.':
            lexeme = self.read_ID_selector()
            tok = Token(TokenType.TYPE_SELECTOR, lexeme)
        elif lexeme == '#':
            lexeme = self.read_ID_selector()
            tok = Token(TokenType.ID_SELECTOR, lexeme)
        elif lexeme == '>':
            tok = Token(TokenType.CHILDREN_SELECTOR, lexeme)
        elif lexeme == '%':
            raise NotImplementedError()

        elif is_letter(lexeme):
            if is_color(lexeme):
                pass
            elif self.is_property():
                lexeme = self.read_property()
                tok = Token(TokenType.PROPERTY, lexeme)
            elif self.is_keyword():
                tok = self.read_keyword()
            elif self.is_mixin_name():
                lexeme = self.read_identifier()
                tok = Token(TokenType.IDENTIFIER, lexeme)
            else:
                tok = self.read_selector()
        else:
            raise TokenUnKnownException('Token {0} does unknown!!!'.format(lexeme))

        self.read_char()
        return tok

    def read_identifier(self):
        pos = self.cur_pos
        while is_letter(self.peek_char):
            self.read_char()
        return self.input_str[pos:self.peek_pos]

    def read_number(self):
        pos = self.cur_pos
        while is_digit(self.peek_char):  # int
            self.read_char()

        if self.peek_char == '.':  # float
            while is_digit(self.peek_char):
                self.read_char()
        return self.input_str[pos:self.cur_pos + 1]

    def is_keyword(self) -> bool:
        pos = self.cur_pos
        while is_letter(self.read_char()):
            pass
        lexeme = self.input_str[pos: self.cur_pos]
        self.cur_pos = pos
        self.cur_char = self.input_str[self.cur_pos]
        self.peek_pos = pos + 1
        self.peek_char = self.input_str[self.peek_pos]

        if lookup_keyword(lexeme):
            return True
        return False

    def read_keyword(self):
        pos = self.cur_pos

        while is_letter(self.peek_char):
            self.read_char()
        lexeme = self.input_str[pos:self.peek_pos]
        token_type = lookup_keyword(lexeme)
        if token_type is None:
            raise TokenUnKnownException('Token {0} does unknown!!!'.format(lexeme))
        return Token(token_type, lexeme)

    def skip_white_space(self):
        while is_white_space(self.cur_char):
            self.read_char()

    def read_selector(self):
        pos = self.cur_pos
        token_type = None
        while is_letter(self.read_char()):
            pass

        # 属性选择器
        if self.cur_char == '[':
            while self.cur_char != ']':
                self.read_char()
            self.read_char()  # 读掉 ]
            token_type = TokenType.PROPERTY_SELECTOR
        elif self.cur_char == ':':
            if self.read_char() == ':':
                token_type = TokenType.SUBWIDGET_SELECTOR
            else:
                token_type = TokenType.PRODO_SELECTOR
            while is_letter(self.read_char()):
                pass
        else:
            token_type = TokenType.CLASS_SELECTOR
        return Token(token_type, self.input_str[pos:self.cur_pos])

    def read_ID_selector(self):
        pos = self.cur_pos
        self.read_char()
        while is_letter(self.peek_char):
            self.read_char()
        return self.input_str[pos:self.cur_pos + 1]

    def is_property(self) -> bool:
        pos = self.cur_pos
        while is_letter(self.read_char()):
            pass
        lexeme = self.input_str[pos: self.cur_pos]
        self.cur_pos = pos
        self.cur_char = self.input_str[self.cur_pos]
        self.peek_pos = pos + 1
        self.peek_char = self.input_str[self.peek_pos]

        if is_property(lexeme):
            return True
        return False

    def read_property(self) -> str:
        pos = self.cur_pos
        while is_letter(self.peek_char):
            self.read_char()
        lexeme = self.input_str[pos: self.peek_pos]
        if is_property(lexeme):
            return lexeme
        self.cur_pos = pos
        self.cur_char = self.input_str[pos]

    def skip_comment(self):
        if self.cur_char == '/':
            if self.peek_char == '/':
                while self.cur_char != '\n':
                    self.read_char()
                self.read_char()  # skip \n
            elif self.peek_char == '*':
                stack = 1
                while stack != 0:
                    self.read_char()
                    if self.cur_char == '/' and self.peek_char == '*':
                        stack += 1
                    elif self.cur_char == '*' and self.peek_char == '/':
                        stack -= 1
                # skip */
                self.read_char()
                self.read_char()

    def is_mixin_name(self):
        i = 0
        while is_letter(self.input_str[self.cur_pos + i]):
            i += 1
        while is_white_space(self.input_str[self.cur_pos + i]):
            i += 1
        ch = self.input_str[self.cur_pos + i]
        return ch == '('



import abc

from pqss.lex import TokenType, Token
from pqss.env import *


class Node(abc.ABC):
    @abc.abstractmethod
    def token_literal(self) -> str:
        pass

    @abc.abstractmethod
    def eval(self, environment: Environment):
        pass


class Statement(Node):
    @abc.abstractmethod
    def stmt_node(self):
        pass


class Expression(Node):
    @abc.abstractmethod
    def expr_node(self):
        pass


class StyleSheet(Node):
    def eval(self, environment: Environment):
        qss = ''
        for stmt in self.statements:
            result = stmt.eval(environment)
            if result is not None:
                qss = qss + str(result)

        return qss

    def __init__(self):
        self.statements: list[Statement] = []

    def token_literal(self):
        if len(self.statements) > 0:
            return self.statements[0].token_literal()
        else:
            return ''


class VarStatement(Statement):
    def eval(self, environment: Environment):
        environment.set(self.name.token.literal,
                        self.value.eval(environment))

    def __init__(self):
        self.token: Token | None = None  # TODO May be not needed
        self.name: Identifier | None = None
        self.value: Expression | None = None

    def token_literal(self) -> str:
        return self.token.literal

    def stmt_node(self):
        pass


class Identifier(Expression):
    def eval(self, environment: Environment):
        val = environment.get(self.token.literal)
        if val:
            return val
        else:
            # environment.set(self.token.literal, self.value)
            return self.value

    def __init__(self, token: Token, value):
        self.token: Token | None = token
        self.value: str = ''

    def token_literal(self) -> str:
        return self.token.literal

    def expr_node(self):
        pass


class ExpressionStatement(Statement):

    def eval(self, environment: Environment):
        return self.expr.eval(environment)

    def __init__(self):
        self.token: Token | None = None
        self.expr: Expression | None = None

    def token_literal(self) -> str:
        return self.token.literal

    def stmt_node(self):
        pass


class IntegerLiteral(Expression):
    def expr_node(self):
        pass

    def eval(self, environment: Environment):
        return self.value

    def token_literal(self) -> str:
        return self.token.literal

    def __init__(self, token: Token | None = None, value: float | None = None):
        self.token = token
        self.value = value


class PrefixExpression(Expression):
    def eval(self, environment: Environment):
        right = self.right.eval(environment)
        if self.token.token_type == TokenType.SUB:
            return -right

    def __init__(self, token: Token | None, operator: str | None = None, right: Expression | None = None):
        self.token = token
        self.operator = operator
        self.right = right

    def expr_node(self):
        pass

    def token_literal(self) -> str:
        pass


class InfixExpression(Expression):
    def eval(self, environment: Environment):
        left = self.left.eval(environment)
        right = self.right.eval(environment)
        if self.token.token_type == TokenType.PLUS:
            return left + right
        if self.token.token_type == TokenType.SUB:
            return left - right
        if self.token.token_type == TokenType.MUL:
            return left * right
        else:
            # self.token.token_type == TokenType.DIV:
            return left / right

    def __init__(self, token: Token, operator: str, left: Expression = None, right: Expression = None):
        self.token = token
        self.left = left
        self.operator = operator
        self.right = right

    def expr_node(self):
        pass

    def token_literal(self) -> str:
        pass


class Color(Expression):
    def eval(self, environment: Environment):
        return self.token.literal

    def __init__(self):
        self.token: Token | None = None

    def expr_node(self):
        pass

    def token_literal(self) -> str:
        pass


class HexColor(Expression):
    def eval(self, environment: Environment):
        return self.token.literal

    def __init__(self):
        self.token: Token | None = None

    def expr_node(self):
        pass

    def token_literal(self) -> str:
        pass


class Selector(Statement):
    def eval(self, environment: Environment):
        return self.token.literal

    def __init__(self, token: Token | None):
        self.token = token

    def stmt_node(self):
        pass

    def token_literal(self) -> str:
        pass


class Rule(Statement):
    def eval(self, environment: Environment):
        res = ''
        res += self.property.literal
        res += ':'
        res += str(self.value.eval(environment))
        res += ';'
        return res

    def __init__(self):
        self.property: Token | None = None
        self.value: Expression | None = None

    def stmt_node(self):
        pass

    def token_literal(self) -> str:
        pass


class Ruleset(Statement):
    def eval(self, environment: Environment):
        res = ''
        selectors = ''
        for selector in self.selector_list:
            selectors += ' ' + selector.eval(environment)
        res += selectors
        res += '{'
        for rule in self.rule_list:
            res += rule.eval(environment)
        res += '}'

        def eval_child(sec, children):
            res = ''
            s = ''
            for se in sec:
                s += ' ' + se.eval(environment)
            s = s.replace(' ', '')
            for child_ruleset in children:
                val = child_ruleset.eval(environment)

                if val.find('&') != -1:
                    val = val.replace('&', s.split(' ')[-1])
                    val = val.replace(' ', '')
                    s = s.split(s.split(' ')[-1])[0]
                    res += s + val
                else:
                    res += s + ' '
                    res += val

            return res

        child_rules = eval_child(self.selector_list, self.child_ruleset)

        res += child_rules
        return res

    def __init__(self):
        self.selector_list: list[Selector] | None = []
        self.rule_list: list[Rule] = []
        self.child_ruleset: list[Ruleset] = []

    def stmt_node(self):
        pass

    def token_literal(self) -> str:
        pass


class Boolean(Expression):
    def eval(self, environment: Environment):
        return self.value

    def expr_node(self):
        pass

    def token_literal(self) -> str:
        pass

    def __init__(self, token: Token, value: bool):
        self.token = token
        self.value = value


class BlockStatement(Statement):
    def eval(self, environment: Environment):
        res = '{'
        for stmt in self.statements:
            res += str(stmt.eval(environment))
        res += '}'
        return res

    def __init__(self, token: Token | None):
        self.token = token
        self.statements: list[Statement] = []

    def stmt_node(self):
        pass

    def token_literal(self) -> str:
        pass


class IfStatement(Statement):
    def eval(self, environment: Environment):
        pass

    def __init__(self, token: Token, expr: Expression | None = None,
                 consequence: BlockStatement | None = None,
                 alternative: BlockStatement | None = None):
        self.token = token
        self.condition = expr
        self.consequence = consequence
        self.alternative = alternative

    def stmt_node(self):
        pass

    def token_literal(self) -> str:
        pass


class Mixin(Statement):
    def eval(self, environment: Environment):
        pass

    def stmt_node(self):
        pass

    def token_literal(self) -> str:
        pass

    def __init__(self, token: Token, name: str | None = None):
        self.token = token
        self.name: str = name
        self.params: list[Identifier] = []
        self.body: BlockStatement | None = None


class Include(Statement):
    def eval(self, environment: Environment):
        pass

    def __init__(self, token: Token, mixin_name: str | None = None, args: list[Expression] | None = None):
        self.token = token
        self.mixin_name = mixin_name
        self.args = args

    def stmt_node(self):
        pass

    def token_literal(self) -> str:
        pass



from enum import Enum, auto


class TokenType(Enum):
    ILLEGAL = auto()  # unknown token
    EOF = auto()  # end of file

    # math
    ASSIGN = auto()  # :
    PLUS = auto()  # +
    SUB = auto()  # -
    MUL = auto()  # *
    DIV = auto()  # /
    EQ = auto()
    NOT_EQ = auto()
    LT = auto()
    GT = auto()

    TRUE = auto()
    FALSE = auto()

    IF = auto()
    ELSE = auto()

    IDENTIFIER = auto()  # id
    SEMICOLON = auto()  # ;
    COMMA = auto()

    LEFT_PAREN = auto()  # (
    RIGHT_PAREN = auto()  # )
    LEFT_BRACE = auto()  # {
    RIGHT_BRACE = auto()  # ｝

    NUMBER = auto()  # 1, 12, 3.0, 3.14
    STRING = auto()
    # 选择器   或许选择器不用分开编写， 可语义分析时候还是要用的
    GENERAL_SELECTOR = auto()  # *
    TYPE_SELECTOR = auto()  # .MyButton
    CLASS_SELECTOR = auto()  # QPushButton
    ID_SELECTOR = auto()  # #btn
    PROPERTY_SELECTOR = auto()  # QPushButton[name='abc']
    CHILDREN_SELECTOR = auto()  # >
    GROUP_SELECTOR = auto()  # #a, #b
    PARENT_REFERENCE = auto()
    # POSTERITY_SELECTOR = auto()
    PRODO_SELECTOR = auto()  # 伪类选择器 :hover
    SUBWIDGET_SELECTOR = auto()  # 子组件选择器 ::indicator
    COLOR = auto()

    # keywords
    IMPORT = auto()  # @import
    DEFAULT = auto()
    EXTEND = auto()
    MIXIN = auto()
    INCLUDE = auto()

    PROPERTY = auto()


keywords = {
    '@import': TokenType.IMPORT,
    '!default': TokenType.DEFAULT,
    '@extend': TokenType.EXTEND,
    '@mixin': TokenType.MIXIN,
    '@include': TokenType.INCLUDE,
    '@if': TokenType.IF,
    '@else': TokenType.ELSE,
    'true': TokenType.TRUE,
    'false': TokenType.FALSE
}

color = ['red', 'blue', 'white', 'yellow']
properties = ['left', 'width', 'background-color']


def is_property(lexeme: str):
    return lexeme in properties


def is_property_prefix(lexeme: str):
    return False


def is_color(lexeme: str):
    return False


def lookup_keyword(lexeme):
    return keywords.get(lexeme)


class Token:
    def __init__(self, token_type: TokenType | None, lexeme):
        self.token_type = token_type
        self.literal = lexeme
